
from regexp_sar import _regexp_sar_c


class RegexpSar(object):
    def __init__(self):
        super().__init__()

        # NOTE: _internal_sar_object is automatically cleaned when has no more references
        self._internal_sar_object = _regexp_sar_c.create_object()

    def add_regexp(self, regexp, callback):
        _regexp_sar_c.add_regexp(self._internal_sar_object, regexp, callback)

    def match_from(self, text, from_position):
        _regexp_sar_c.match_from(self._internal_sar_object, text, from_position)

    def match_at(self, text, at_position):
        _regexp_sar_c.match_at(self._internal_sar_object, text, at_position)

    def match(self, text):
        self.match_from(text, 0)

    def continue_from(self, from_pos):
        _regexp_sar_c.continue_from(self._internal_sar_object, from_pos)

    def stop_match(self):
        _regexp_sar_c.stop_match(self._internal_sar_object)



from regexp_sar import RegexpSar

# create a SAR instance
sar = RegexpSar()

# initialize lists of all the sentences positions of alice with cat/rabbit
alice_and_cat_position = []
alice_and_rabbit_position = []

# keep track of sentence start position
sentence_start_position = 0

# flags for whether alice/cat/rabbit occurs in the current sentence
is_alice_in_sentence = False
is_cat_in_sentence = False
is_rabbit_in_sentence = False

# mark that the word alice is found in the current sentence


def find_alice(f, t):
    global is_alice_in_sentence
    is_alice_in_sentence = True


# mark that the word cat is found in the current sentence
def find_cat(f, t):
    global is_cat_in_sentence
    is_cat_in_sentence = True


# mark that the word rabbit is found in the current sentence
def find_rabbit(f, t):
    global is_rabbit_in_sentence
    is_rabbit_in_sentence = True


# indicates that the sentece is over,
# append sentence into the right list if found any
def end_of_sentence(from_pos, to_pos):
    global is_cat_in_sentence, is_alice_in_sentence, is_rabbit_in_sentence
    global alice_and_cat_position, alice_and_rabbit_position
    global sentence_start_position

    # check if alice is found with cat/rabbit in current sentence
    if is_alice_in_sentence and is_cat_in_sentence:
        alice_and_cat_position.append([sentence_start_position, to_pos])
    if is_alice_in_sentence and is_rabbit_in_sentence:
        alice_and_rabbit_position.append([sentence_start_position, to_pos])

    # set next sentence start position to end of current sentence position + 1
    sentence_start_position = to_pos + 1

    # reset words in sentence to false
    is_alice_in_sentence = False
    is_cat_in_sentence = False
    is_rabbit_in_sentence = False


# get Alice In Wonderland book content
with open("./alice_in_wonderland.txt", "r") as alice_book:
    text = alice_book.read()

# add alice/cat/rabbit regexps
sar.add_regexp('\\^\\walice\\^\\w', find_alice)
sar.add_regexp('\\^\\wAlice\\^\\w', find_alice)
sar.add_regexp('\\^\\wcat\\^\\w', find_cat)
sar.add_regexp('\\^\\wCat\\^\\w', find_cat)
sar.add_regexp('\\^\\wrabbit\\^\\w', find_rabbit)
sar.add_regexp('\\^\\wRabbit\\^\\w', find_rabbit)

# add end of sentence regexps,
# NOTE that they point to the same callback
for cur_re in ['\\.', '\\?', '!', ';']:
    sar.add_regexp(cur_re, end_of_sentence)

# run the regexps on the text
sar.match(text)

# the matching has completed, show the results:
print("Alice + Cat Matches: " + str(alice_and_cat_position))
print("Alice + Rabbit Matches: " + str(alice_and_rabbit_position))



from regexp_sar import RegexpSar

sar = RegexpSar()

match_word_count = 0
match_word = None
match_num_count = 0
match_num = None
match_str = "hello world 123 abc 456 789"

regexps = [
    ['\w+', 'word'],
    ['\d+', 'number'],
]

for cur_regexp in regexps:
    def find_second_match(description):
        match_count = 0
        match_val = None
        def callback(from_pos, to_pos):
            nonlocal match_count, match_val
            match_count += 1
            if match_count == 2:
                print("Match: " + str(description) + ": " + match_str[from_pos:to_pos])
            sar.continue_from(to_pos)
        return callback
    sar.add_regexp(cur_regexp[0], find_second_match(cur_regexp[1]))

'''
Output:
    Match: word: world
    Match: number: 456
'''
sar.match(match_str)



from regexp_sar import RegexpSar

sar = RegexpSar()


match_str = "hello world abc"

sar.add_regexp('\w+', lambda from_pos,
               to_pos: print("Match Word: " + match_str[from_pos:to_pos]))
sar.add_regexp('world', lambda from_pos, to_pos: print(
    "Found world from: " + str(from_pos) + " to: " + str(to_pos)))

'''
Output:
    Match Word: hello
    Match Word: ello
    Match Word: llo
    Match Word: lo
    Match Word: o
    Match Word: world
    Found world from: 6 to: 11
    Match Word: orld
    Match Word: rld
    Match Word: ld
    Match Word: d
    Match Word: abc
    Match Word: bc
    Match Word: c
'''
sar.match(match_str)




from nose.tools import nottest

from regexp_sar import RegexpSar


class VarPointer:
    def __init__(self):
        super().__init__()
        self.val = 0

    def set(self, val):
        self.val = val

    def append(self, amount):
        self.val += amount


@nottest
def validate(expected_val, calculated_val, msg=None):
    if msg is None:
        msg = f"expected: {expected_val}, received: {calculated_val}"
    assert expected_val == calculated_val, msg


def test1():
    obj = VarPointer()
    sar = RegexpSar()
    sar.add_regexp("abcd", lambda from_pos, to_pos: obj.append(1))
    sar.match("qabcdef")
    validate(1, obj.val)


def test2():
    obj = VarPointer()
    sar = RegexpSar()
    sar.add_regexp("abcd", lambda from_pos, to_pos: obj.append(1))
    sar.match("qabcde abcd f")
    validate(2, obj.val)


def test3():
    obj = VarPointer()
    sar = RegexpSar()
    sar.add_regexp("abcd", lambda from_pos, to_pos: obj.append(1))
    sar.match("qqqqqqqqqqqq")
    validate(0, obj.val)


def test4():
    obj = VarPointer()
    sar = RegexpSar()
    sar.add_regexp("abcd", lambda from_pos, to_pos: obj.append(1))
    sar.add_regexp("nm", lambda from_pos, to_pos: obj.append(2))
    my_str = "qabcdefnmq"
    sar.match_from(my_str, 1)
    validate(3, obj.val)

    obj.set(0)
    sar.match_from(my_str, 2)
    validate(2, obj.val)


def test5():
    obj = VarPointer()
    sar = RegexpSar()
    re = "abcd"
    myStr = "qabcdef"
    sar.add_regexp(re, lambda from_pos, to_pos: obj.append(1))
    sar.match(myStr)
    validate(1, obj.val)


def test6():
    obj = VarPointer()
    sar = RegexpSar()
    myStr = "0123456 789"
    sar.add_regexp("\\d+", lambda from_pos,
                   to_pos: obj.set(myStr[from_pos:to_pos]))
    sar.match_at(myStr, 3)
    validate("3456", obj.val)


def test7():
    obj = VarPointer()
    sar = RegexpSar()
    sar.add_regexp("abcd", lambda from_pos, to_pos: obj.append(1))
    sar.match("qabcdeabcdkabcdf")
    validate(3, obj.val)

    obj2 = VarPointer()
    sar2 = RegexpSar()

    def sar2_callback(from_pos, to_pos):
        obj2.append(1)
        if obj2.val == 2:
            sar2.stop_match()

    sar2.add_regexp("abcd", sar2_callback)
    sar2.match("qabcdeabcdkabcdf")
    validate(2, obj2.val)


def test8():
    objFrom = VarPointer()
    objTo = VarPointer()
    sar = RegexpSar()
    myStr = "qabcdef"
    re = "abcd"

    def sar_callback(from_pos, to_pos):
        objFrom.set(from_pos)
        objTo.set(to_pos)

    sar.add_regexp(re, sar_callback)
    sar.match(myStr)
    validate(1, objFrom.val)
    validate(5, objTo.val)
    matchStr = myStr[objFrom.val:objTo.val]
    validate(re, matchStr)
    objFrom.set(0)
    objTo.set(0)
    sar.match("qqqqabcdttt")
    validate(4, objFrom.val)
    validate(8, objTo.val)


def test9():
    objFrom = VarPointer()
    objTo = VarPointer()
    sar = RegexpSar()
    myStr = "1234"
    re = "\\d+"

    def sar_callback(from_pos, to_pos):
        objFrom.set(from_pos)
        objTo.set(to_pos)
        sar.stop_match()

    sar.add_regexp(re, sar_callback)
    sar.match(myStr)
    validate(myStr, myStr[objFrom.val:objTo.val])


def test10():
    myStr = "abc"
    res = []
    sar = RegexpSar()
    sar.add_regexp("\\w+",
                   lambda from_pos, to_pos: res.append(myStr[from_pos:to_pos]))
    sar.match(myStr)
    validate("abc", res[0])
    validate("bc", res[1])
    validate("c", res[2])
    validate(3, len(res))


def test11():
    myStr = "123abc456"
    res = []
    sar = RegexpSar()

    def sar_callback(from_pos, to_pos):
        res.append(myStr[from_pos:to_pos])
        sar.continue_from(to_pos + 1)

    sar.add_regexp("\\a+", sar_callback)
    sar.match(myStr)
    validate("abc", res[0])
    validate(1, len(res))


def test12():
    myStr = "123abc456"
    res = []
    sar = RegexpSar()

    def call(from_pos, to_pos):
        res.append(myStr[from_pos:to_pos])
        sar.continue_from(to_pos)

    sar.add_regexp("\\a+", call)
    sar.add_regexp("\\d+", call)
    sar.match(myStr)

    validate("123", res[0])
    validate("abc", res[1])
    validate("456", res[2])
    validate(3, len(res))


def test13():
    myStr = "123abc"
    res = []
    sar = RegexpSar()

    def call(from_pos, to_pos):
        res.append(myStr[from_pos:to_pos])
        sar.continue_from(to_pos)

    sar.add_regexp("\\d+", call)
    sar.add_regexp("\\w+", call)
    sar.match(myStr)

    validate("123", res[0])
    # TODO check if this is a real error?
    validate("123abc", res[1])
    validate(2, len(res))


def test14():
    myStr = "backup:x:34:34:backup:/var/backups:/usr/sbin/nologin"
    elemNum = 6
    sar = RegexpSar()
    matchCount = VarPointer()
    strt = VarPointer()
    end = VarPointer()

    def sar_callback(from_pos, to_pos):
        matchCount.append(1)
        sar.continue_from(to_pos)
        if matchCount.val == elemNum - 1:
            strt.set(to_pos)
        elif matchCount.val == elemNum:
            end.set(from_pos)
            sar.stop_match()

    sar.add_regexp(":", sar_callback)
    sar.match(myStr)
    if end.val > 0:
        matchStr = myStr[strt.val:end.val]
        validate("/var/backups", matchStr)
    else:
        # fail on purpuse
        validate(0 == 1, "no match found")

# TODO: check what this test is for?


def test15():
    myStr = "123abc"
    obj = VarPointer()
    sar = RegexpSar()

    def call(from_pos, to_pos):
        obj.set(1)

    sar.add_regexp("\\d+", call)
    sar.add_regexp("\\w+", call)
    sar.match(myStr)

    validate(1, obj.val)


def test16():
    myStr = "123abc"
    res = []
    sar = RegexpSar()

    def call(from_pos, to_pos):
        res.append(myStr[from_pos:to_pos])
        sar.continue_from(to_pos + 1)

    sar.add_regexp("\\d+", call)
    sar.add_regexp("\\a+", call)
    sar.match(myStr)

    validate("123", res[0])
    validate("bc", res[1])
    validate(2, len(res))


def test17():
    myStr = "mm abc nn"
    obj = VarPointer()
    sar = RegexpSar()

    def call(from_pos, to_pos):
        obj.set(1)

    sar.add_regexp("\\d+", call)
    sar.add_regexp("\\a+", call)
    sar.match(myStr)

    if obj.val != 1:
        validate(0 == 1, "regexp should match")


def test18():
    myStr = "first abbbbc second 123 end"
    matches = []
    regexps = [
        ["ab+c", "First Match"],
        ["\\d+", "Second Match"],
    ]

    sar = RegexpSar()
    for re in regexps:
        reStr = re[0]
        reTitle = re[1]

        def gen_call(reTitleInternal):
            def call(from_pos, to_pos):
                matchStr = myStr[from_pos:to_pos]
                matches.append(reTitleInternal + ": " + matchStr)
                sar.continue_from(to_pos)
            return call

        sar.add_regexp(reStr, gen_call(reTitle))

    sar.match(myStr)
    validate(2, len(matches))
    validate("First Match: abbbbc", matches[0])
    validate("Second Match: 123", matches[1])


def test19():
    # get third match and stop
    myStr = "aa11 bb22 cc33 dd44"
    matchCount = VarPointer()
    matchedStr3 = VarPointer()
    sar = RegexpSar()

    def call(from_pos, to_pos):
        matchCount.append(1)
        if matchCount.val == 3:
            matchedStr3.val = myStr[from_pos:to_pos]
            sar.stop_match()
        else:
            sar.continue_from(to_pos)

    sar.add_regexp("\\w+", call)
    sar.match(myStr)

    validate(3, matchCount.val)
    validate("cc33", matchedStr3.val)


def test20():
    # get match only at certain position
    myStr = "aa11 bb22 cc33 dd44"
    matchedStr4 = VarPointer()
    sar = RegexpSar()

    def call(from_pos, to_pos):
        matchedStr4.val = myStr[from_pos:to_pos]
    sar.add_regexp("\\w+", call)
    sar.match_at(myStr, 5)

    validate("bb22", matchedStr4.val)


def test21():
    # get match only at certain position
    myStr = "aaaaaaaaaaaaaaaaaaaaaaaaaabbbc"
    matched = VarPointer()
    sar1 = RegexpSar()
    sar2 = RegexpSar()

    def sar1_callback(from_pos, to_pos):
        sar2.match_at(myStr, to_pos)
        sar1.stop_match()

    sar2.add_regexp("b+c", lambda from_pos, to_pos: matched.append(1))
    sar1.add_regexp("a+", sar1_callback)
    sar1.match(myStr)

    validate(1, matched.val)


def test22():
    sar = RegexpSar()
    alpha_pos = VarPointer()
    alpha_pos.val = -1
    anchorPos = VarPointer()

    def sar_digit_callback(digit_pos, to_pos):
        if alpha_pos.val != -1:
            dist = digit_pos - alpha_pos.val
            if (dist == 1):
                anchorPos.set(digit_pos)

    sar.add_regexp("\\a", lambda from_pos, to_pos: alpha_pos.set(from_pos))
    sar.add_regexp("\\d", sar_digit_callback)
    sar.match("aa bb2cc dd")
    validate(5, anchorPos.val)



import random
import re
import time
from regexp_sar import RegexpSar

# generates a random string of length "size"
def str_gen(size):
    return "".join(random.choices("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890", k=size))

# list contains all the regexps
regexp_list = []
print("Gen regexps")
# define each regexp length
re_len = 5

# define amount of regexps required
regexp_amount = 10_000

for i in range(regexp_amount):
    cur_re = str_gen(re_len)
    while cur_re in regexp_list:
        cur_re = str_gen(re_len)
    regexp_list.append(cur_re)
print("Done gen regexps")

print("Gen match string")
# generate the text on which we will run
match_str = str_gen(1_000_000)
print("Done gen match string")

# counts amount of matches for SAR matchingg
match_count = 0
# this will be called every time any regexp was matched in the text
def count_matches(from_pos, to_pos):
    global match_count
    match_count += 1


# benchmark native Python regexp engine performance
print("Start python match")
# generate Python regexp string using alternation
python_native_re = f"({'|'.join(regexp_list)})"
start = time.time()

# find amount of matches found with native Python regexp engine
python_re_match_count = len(re.findall(python_native_re, match_str))
print(f"Done in {time.time() - start} seconds {python_re_match_count} matches")


# create a SAR instance
sar = RegexpSar()

# append all the regexps for the SAR instance
for cur_re in regexp_list:
    sar.add_regexp(cur_re, count_matches)

# benchmark SAR performance
print("Start SAR Match")
start = time.time()
sar.match(match_str)
print(f"Done Match in: {time.time() - start} seconds. {match_count} matches!")




from regexp_sar import RegexpSar

# Command to be evaluated
command = "(1 + 1   ) ^ (4 - (2 - 1))"


class StateSar:
    def __init__(self, init_state):
        super().__init__()
        self._sar = RegexpSar()
        self._state = init_state
        self._all_states = set()

    def add_regexp(self, state, regexp, regexp_callback):
        self._all_states.add(state)

        def inner_callback(from_pos, to_pos):
            if self._state == state:
                regexp_callback(from_pos, to_pos)
            self._sar.continue_from(to_pos)
        self._sar.add_regexp(regexp, inner_callback)

    def match(self, match_str):
        self._sar.match(match_str)

    def set_state(self, new_state):
        if new_state not in self._all_states:
            raise Exception("Unknown state: " + str(new_state))
        self._state = new_state

    def get_state(self):
        return self._state


stack = [[]]

state_sar = StateSar("number")


def handle_number(from_pos, to_pos):
    match_str = command[from_pos:to_pos]
    stack[-1].append(float(match_str))
    state_sar.set_state("action")


def handle_action(from_pos, to_pos):
    match_str = command[from_pos:to_pos]
    stack[-1].append(match_str)
    state_sar.set_state("number")


def handle_paren(from_pos, to_pos):
    stack.append([])
    state_sar.set_state("number")


def close_paren(from_pos, to_pos):
    cur_stack = stack.pop()
    stack[-1].append(cur_stack)
    state_sar.set_state("action")


def eval_calc(expr):
    if type(expr) is float:
        return expr

    # expr is list!
    actions = {
        "+": lambda a, b: a + b,
        "-": lambda a, b: a - b,
        "*": lambda a, b: a * b,
        "/": lambda a, b: a / b,
        "^": lambda a, b: a ** b,
    }
    clone_list = expr[:]
    while True:
        arg1_expr = clone_list.pop(0)
        if len(clone_list) == 0:
            return eval_calc(arg1_expr)
        # we ensured before if it contains an action, it'll contain another arg
        cur_action = clone_list.pop(0)
        arg2_expr = clone_list.pop(0)

        clone_list.insert(0, actions[cur_action](
            eval_calc(arg1_expr), eval_calc(arg2_expr)))


state_sar.add_regexp("number", '\d+', handle_number)

action_chars = "+*-/^"
for cur_action_char in action_chars:
    state_sar.add_regexp("action", cur_action_char, handle_action)

state_sar.add_regexp("number", "(", handle_paren)
state_sar.add_regexp("action", ")", close_paren)


state_sar.match(command)
if state_sar.get_state() != "action":
    raise Exception(f"Command is incomplete! missing {state_sar.get_state()}")
if len(stack) > 1:
    raise Exception(
        f"Command is incomplete! didnt close {len(stack)-1} parens")
stack = stack[0]

print("-------")
print(command)
print(stack)
print(f"val = {eval_calc(stack)}")



from setuptools import setup, Extension


def main():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
    setup(name="regexp-sar",
          version="0.1.2b5",

          description="SAR Regexp Engine for Python",
          long_description=long_description,
          long_description_content_type="text/markdown",

          url="https://github.com/nmnsnv/regexp_sar",

          author="Noam Nisanov",
          author_email="noam.nisanov@gmail.com",

          test_suite='nose.collector',
          tests_require=['nose'],

          ext_modules=[Extension("regexp_sar._regexp_sar_c",
                                 ["src/sarmodule.c", "src/sarcore.c"],
                                 py_limited_api=True)],
          headers=['src/sarcore.h'],
          include_dirs=["src/"],
          license="MIT",
          python_requires='>=3.4',
          packages=["regexp_sar"])


if __name__ == "__main__":
    main()




from nose.tools import nottest

from regexp_sar import RegexpSar


def test_matching():
    match_test(["abcd"], "qabcdef", 1)
    match_test(["abcd", "abcf"], "zqabcdwabcfr", 3)
    match_test(["abcd", "abc"], "zqabcdwr", 3)

    match_test(["abcd", "amk", "axf", "awe", "aitr"], "zqabcdwamkraxfaz", 7)
    match_test(["a"], "a", 1)
    match_test(["a"], "b", 0)
    match_test(["a"], "ma", 1)
    match_test(["a"], "maa", 2)
    match_test(["asv"], "bbbbasvmm", 1)
    match_test(["asv"], "asv", 1)
    match_test(["asv", "kdf"], "bbasvmkdfm", 3)

    match_test(["as", "ask"], "bbasvmaskm", 4)
    match_test(["as", "ask"], "bbaskmmmm", 3)

    match_test(["ab?cd", "dk?m"], "aaadmabcdn", 3)
    match_test(["ab?cd", "k?m"], "aaadmabcdn", 3)
    match_test(["zzm?"], "iozzj", 1)
    match_test(["zzm?"], "iozzmj", 2)
    match_test(["zzm?n?"], "iozzjzznj", 3)
    match_test(["ab?cde", "acdf"], "mmabcdfkkacdeoo", 1)
    match_test(["ab?cd?e"], "mabcdem macdem macem mabcem nbcden", 4)
    match_test(["ab?c?d?e"], "maem madem mabem", 3)
    match_test(["ac?c?c?e"], "m ace m", 1)
    match_test(["ac?c?c?e", "ace"], "m ace m", 3)
    match_test(["ac?c?c?c?e"], "m acce m", 1)

    match_test(["ab?cd", "acd"], "qacdq", 3)

    match_test(["ab+cd"], "aadmabcdn", 1)
    match_test(["ab+cd"], "aadmabbcdn", 1)
    match_test(["ab+cd"], "aadmabbbbbbbbcdn", 1)

    match_test(["b+"], "abbbc", 3)
    match_test(["b+c"], "abbbcd", 3)

    match_test(["ab+"], "aadmabbbbcdn", 1)
    match_test(["ab+c+d"], "aadmabcccccdn", 1)
    match_test(["ab+c+d"], "aadmabdn", 0)
    match_test(["ab+b?d"], "aadmabdn", 1)
    match_test(["ab+b?d"], "aadmabbbbdn", 1)
    match_test(["ab+c?d"], "aadmabbbbdn", 1)
    match_test(["ab+b+d"], "aadmabbbbdn", 1)
    match_test(["ab+b+d"], "aadmabdn", 0)
    match_test(["ab+cd", "ab?cd"], "aaacdabbbbbbcdn", 3)

    match_test(["ab*cd"], "aaadmabcdn", 1)
    match_test(["ab*cd"], "aaadmabbbbcdn", 1)
    match_test(["ab*cd"], "aaadmacdn", 1)
    match_test(["ab*c*d"], "aamadn", 1)
    match_test(["ab*c*d"], "aamacccccccdn", 1)

    match_test(["ab*c?d"], "aamadn", 1)
    match_test(["ab*c?d"], "aamabcdn", 1)
    match_test(["ab*c?d"], "aamacdn", 1)
    match_test(["ab*c?d"], "aamabdn", 1)

    match_test(["ab?cd", "ab+cd"], "mmabcdmm", 3)
    match_test(["ab?cd", "ab+cd"], "mmacdmm", 1)
    match_test(["ab?cd", "ab+cd"], "mmabbbcdmm", 2)
    match_test(["ab*c?b+e"], "mmabedmm", 1)
    match_test(["ab*c?b+e"], "mmabcbedmm", 1)
    match_test(["ab+b?b+c"], "mmabbcdmm", 1)
    match_test(["ab+c?b+c"], "mmabbbcdmm", 1)

    match_test(["ab\\cd"], "q1abcdef", 1)
    match_test(["ab\\cd"], "q1abcdef", 1)
    match_test(["ab\\\\cd"], "q2abcdef", 0)
    match_test(["ab\\\\cd"], "q3ab\\cdef", 1)
    match_test(["ab\\\\\\\\cd"], "q3ab\\\\cdef", 1)
    match_test(["ab\\?cd"], "qabcdef", 0)
    match_test(["ab\\?cd"], "qab?cdef", 1)

    match_test(["abcd\\?"], "qabcd?ef", 1)
    match_test(["abcd\\"], "qabcdef", 1)
    match_test(["ab?cd"], "qabcdef", 1)
    match_test(["abcd?"], "qabcef", 1)
    match_test(["abcd?"], "qabcdef", 2)
    match_test(["ab?cd"], "qacdef", 1)
    match_test(["ab?cd"], "qab?cdef", 0)
    match_test(["ab\\?cd"], "qab?cdef", 1)
    match_test(["ab?cd", "abce"], "qacem", 0)
    match_test(["abd", "ab?d", "kkk", "abc?d"], "qabdm", 11)
    match_test(["ab?b?b?b?b?cd"], "qacdef", 1)
    match_test(["ab?b?b?cd"], "qabcdef", 1)
    match_test(["ab?b?b?cd"], "qabbcdef", 1)
    match_test(["abc??"], "qabc?def", 1)

    match_test(["ab+cd"], "qabbbbcdef", 1)
    match_test(["ab+cd"], "qabcdq qabbbbcdq", 2)
    match_test(["am+e"], "qammmeq", 1)
    match_test(["ab+c", "am+e"], "qabcq qabbbbbcq qammmeq", 4)
    match_test(["ab+c+d"], "qabbcdq", 1)

    match_test(["ab+cd", "abcf"], "abbcd", 1)
    match_test(["ab+cd", "abcf"], "abbcf", 0)
    match_test(["ab+cd"], "acdf", 0)
    match_test(["ab+"], "qabcdq qabbbbcdq qacdq", 2)
    match_test(["ab+bcd"], "qabcdq", 0)
    match_test(["ab+bcd"], "qabbcdq", 1)
    match_test(["ab+b+cd"], "qabbcdq", 1)
    match_test(["ab+b+cd"], "qabcdq", 0)
    match_test(["ab+b+b+cd"], "qabbcdq", 0)
    match_test(["ab+b+b+cd"], "qabbbcdq", 1)
    match_test(["ab+b+b+cd"], "qabbbbbbbbbbbbbbbbbcdq", 1)
    match_test(["a\\?+cd"], "qa????cdq", 1)
    match_test(["ab*cd"], "qabbbbcdef", 1)
    match_test(["ab*cd"], "qacdef", 1)

    match_test(["ab?cd+d+ef"], "q acddddef q", 1)
    match_test(["ab?cd+d+ef"], "q acdef q", 0)
    match_test(["ab?cd+d+ef"], "q acddddef q", 1)

    match_test(["a\\++b"], "q a+++++++b q", 1)

    # 3 matches: "bbb", "bb", "b"
    match_test(["b+"], "m bbb m", 3)
    match_test(["ab+"], "m abbb m", 1)

    match_test(["a\\+c"], "m a+c ", 1)
    match_test(["a\\+c"], "m aac ", 0)
    match_test(["ab\\?c"], "m aac ", 0)
    match_test(["ab\\?+c"], "m aab??????c ", 1)

    match_test(["a\\dc", "a5c"], "m a5c m", 3)
    match_test(["a\\dc"], "m a5c a8c m", 2)
    match_test(["ab\\d?c"], "m abc m", 1)
    match_test(["ab\\d?c"], "m ab1c m", 1)
    match_test(["ab\\d?c"], "m abdc m", 0)

    match_test(["ab\\d+c"], "m ab7c m", 1)
    match_test(["ab\\d+c"], "m ab7657234c m", 1)
    match_test(["ab\\d*c"], "m abc m", 1)
    match_test(["ab\\d*c"], "m ab234c m", 1)
    match_test(["ab\\d+\\d+c"], "m ab234c m", 1)
    match_test(["ab\\d+\\d+c"], "m ab2c m", 0)

    match_test(["\\d+\\d+c"], "m ab234c m", 2)
    match_test(["\\d+"], "m ab234c m", 3)

    match_test(["a\\sc", "a.c"], "m a c m", 3)
    match_test(["a\\sc"], "m a c a c m", 2)
    match_test(["ab\\s?c"], "m abc m", 1)
    match_test(["a\\s*c"], "m ac a c a      c m", 3)

    match_test(["\\a"], " t ", 1)
    match_test(["\\a"], " 4 ", 0)
    match_test(["\\a+"], " t ", 1)
    match_test(["\\a+"], " 4 ", 0)
    match_test(["\\a+"], " tre ", 3)
    match_test(["\\a+"], " 456 ", 0)
    match_test(["a\\ac", "a.c"], "m afc m", 3)
    match_test(["a\\ac"], "m asc auc m", 2)
    match_test(["a\\ac"], "m a4c m", 0)
    match_test(["ab\\a?c"], "m abc m", 1)
    match_test(["a\\a*c"], "m ac atc asggfc a8c m", 3)

    match_test(["a\\wc"], "m a4c adc m", 2)
    match_test(["ab\\w?c"], "m abtc m", 1)
    match_test(["a\\w*c"], "m ac ahc asd87sc m", 3)

    match_test(["a.c"], "m a5d m", 0)
    match_test(["a.c"], "m akcm alc l ", 2)
    match_test(["a.?c"], "m akcm ac l ", 2)
    match_test(["a.+c"], "m  abbregtjhbc m", 1)
    match_test(["a.+.+c"], "m  ahbc m", 1)
    match_test(["a.+.+c"], "m  ahc m", 0)
    match_test(["a\\.+c"], "m abbc m", 0)
    match_test(["a\\.+c"], "m a..c m", 1)
    match_test(["a\\.c"], "m a.cm", 1)
    match_test(["a\\.c"], "m ahcm", 0)
    match_test(["."], "m", 1)
    match_test(["\\."], "m", 0)
    match_test(["a\\.?b"], "amb", 0)
    match_test(["a\\.?b"], "ab", 1)
    match_test(["a\\.?b"], "a.b", 1)

    match_test(["a\\^bc"], "m adc m", 1)
    match_test(["a\\^bdc"], "m aqdc m", 1)
    match_test(["a\\^bc"], "m abc m", 0)

    match_test(["a\\^b?dc"], "m adc m", 1)
    match_test(["a\\^b?dc"], "m aqdc m", 1)
    match_test(["a\\^b?bc"], "m abc m", 1)
    match_test(["a\\^b?c"], "m ac m", 1)
    match_test(["a\\^b?\\^m?c"], "m attc m", 1)
    match_test(["a\\^b?\\^m?c"], "m ac m", 1)

    match_test(["a\\^bc"], "m adc a c a6c m", 3)
    match_test(["a\\^b\\^dc"], "m affc m", 1)
    match_test(["a\\^b\\^dc"], "m afdc m", 0)
    match_test(["a\\^b\\^dc"], "m abfc m", 0)
    match_test(["a\\^b\\^dc"], "m affa m", 0)
    match_test(["a\\^\\\\c"], "m adc m", 1)
    match_test(["a\\^\\\\c"], "m a\\c m", 0)

    match_test(["a\\^\\sc"], "m a c m", 0)
    match_test(["a\\^\\sc"], "m adc m", 1)
    match_test(["a\\^\\dc"], "m a6c m", 0)
    match_test(["a\\^\\dc"], "m adc m", 1)
    match_test(["a\\^\\ac"], "m adc m", 0)
    match_test(["a\\^\\ac"], "m a6c m", 1)
    match_test(["a\\^\\wc"], "m a6c m", 0)
    match_test(["a\\^\\wc"], "m adc m", 0)
    match_test(["a\\^\\wc"], "m a c m", 1)

    match_test(["a\\^b", "a\\^c"], "ac", 1)

    match_test(["a\\^b+c"], "m adc m", 1)
    match_test(["a\\^b+c"], "m aduic m", 1)
    match_test(["a\\^b+\\^b+c"], "m adfc m", 1)
    match_test(["a\\^b+dc"], "m adghjdc m", 1)
    match_test(["a\\^b+dc"], "m adghbjdc m", 0)
    match_test(["a\\^b*dc"], "m adc m", 1)
    match_test(["a\\^b*c"], "m atc m", 1)
    match_test(["a\\^b*c"], "m atghhgc m", 1)
    match_test(["a\\^\\d+c"], "m adgdc m", 1)
    match_test(["a\\^\\d+c"], "m adghjc m", 1)
    match_test(["a\\^\\d+c"], "m adgh7jc m", 0)
    match_test(["a\\^\\w+c"], "m a-+-c m", 1)
    match_test(["a\\^\\a+c"], "m a-+-c m", 1)
    match_test(["a\\^\\s+c"], "m a-+-c m", 1)


@nottest
def match_test(all_regexp, match_string, match_expected):
    sar = RegexpSar()
    path_res = 0
    for i in range(len(all_regexp)):
        re_num = 2 ** i

        # NOTE: this is done since re_num is not defined every iteration, but only once
        def gen_callback(cur_re_num):
            def match_callback(from_pos, to_pos):
                nonlocal path_res
                path_res += cur_re_num
            return match_callback
        sar.add_regexp(all_regexp[i], gen_callback(re_num))
    sar.match(match_string)
    assert match_expected == path_res, f"Match fail for [{', '.join(all_regexp)}] in >>{match_string}<<"




from regexp_sar import RegexpSar

with open("./c_code.c", "r") as c_fh:
    c_content = c_fh.read()

required_method = "malloc"

sar = RegexpSar()

last_found_word = None


def found_name(from_pos, to_pos):
    global last_found_word
    last_found_word = c_content[from_pos: to_pos]
    sar.continue_from(to_pos)


sar.add_regexp("\\w+", found_name)

# name of last encountered function
last_found_function = None


def found_function(from_pos, to_pos):
    global last_found_function
    last_found_function = last_found_word


sar.add_regexp("(", found_function)

# name of function we are currently in its body (with its body)
inside_function_name = None
curly_bracket_count = 0


def handle_open_curly_bracket(from_pos, to_pos):
    global curly_bracket_count, inside_function_name
    if curly_bracket_count == 0:
        inside_function_name = last_found_function
    curly_bracket_count += 1


sar.add_regexp("{", handle_open_curly_bracket)


def handle_close_curly_bracket(from_pos, to_pos):
    global curly_bracket_count
    curly_bracket_count -= 1


sar.add_regexp("}", handle_close_curly_bracket)


def handle_required_method_found(from_pos, to_pos):
    print(f"found at: {inside_function_name}")


sar.add_regexp(f"\\^\\w{required_method}\\^\\w", handle_required_method_found)

sar.match(c_content)
